def calc_fit_savgol_filter(
        idata, col_in, col_out, window1, window2, order,
        inner_dates=True, col_int=''
        ):
    r"""Fit using Savitzky-Golay filter and smooth the results.

    Parameters
    ----------
    idata : DataFrame
        Input/Output data.
    col_in : str
        Input column.
    col_out : str
        Output column.
    window1 : timedelta
        1/2 time window to perform the fitting.
    window2 : timedelta
        1/2 time window to perform smoothing.
    order : Series
        Order of the function for each timestamp.
    inner_dates : boolean
        If true date limits are set to first/last valid index, else to extend
        by window size.
    col_int : str, optional
        Intermediate column. If set, the column will be populated with the
        values from the moving fit before the averaging.
    """
    selected = idata[col_in].dropna()
    dates = idata.index.to_series()
    window_max = max(window1, window2)
    window_total = window1 + window2
    # Do not progress if two half-windows are longer than the data length
    if dates[0] + 2*window_max > dates[-1]:
        raise ValueError("Window size is larger than the length of data.")
    # Get date ranges
    time0, time1 = get_savgol_date_limits(
        idata=idata, col_in=col_in, window=window1, inner_dates=inner_dates
    )
    date_s, date_e = get_savgol_dates(
        dates=dates, time0=time0-window1, time1=time1+window1, window=window1
    )
    # Normalisation factor to change timedelta to float
    x_norm = pd.to_timedelta('1 h')
    # Apply moving fit
    interpolated = dates[time0 - window_total : time1 + window_total].apply(
        lambda x:
        savgol_value(
            x=(
                selected.loc[date_s[x] : date_e[x]].index
                - date_s[x] - window1  # centre of the window to become 0.
            ) / x_norm,
            y=selected.loc[date_s[x] : date_e[x]],
            order=order[x],
            x_i=(
                (x - window1 - date_s[x]) / x_norm
            )
        )
        if np.array(selected.loc[date_s[x] : date_e[x]]).size >= 2 else
        selected.loc[date_s[x] : date_e[x]].mean()
        if np.array(selected.loc[date_s[x] : date_e[x]]).size >= 1 else
        np.nan
    )
    if col_int:
        idata.loc[interpolated.index, col_int] = interpolated
    # Apply moving average
    idata[col_out].update(
        dates[time0 - window2 : time1 + window2].apply(
            lambda x: interpolated.loc[x - window2 : x + window2].mean()
        )
    )
    pd.concat([date_s, date_e, interpolated, idata[col_out]], axis=1).to_csv('dates.csv')
    idata[col_out].loc[: selected.first_valid_index()] = (
        idata[col_out].loc[selected.first_valid_index()]
    )
    idata[col_out].loc[selected.last_valid_index() :] = (
        idata[col_out].loc[selected.last_valid_index()]
    )


def get_savgol_date_limits(idata, col_in, window, inner_dates=True):
    r"""Fit using Savitzky-Golay filter and smooth the results.

    Parameters
    ----------
    idata : DataFrame
        Input/Output data.
    col_in : str
        Input column.
    window : timedelta
        1/2 time window to perform the fitting.
    inner_dates : boolean
        If true date limits are set to first/last valid index, else to extend
        by window size.
    """
    selected = idata[col_in].dropna()
    dates = idata.index.to_series()
    null = pd.to_timedelta('0 h')
    # Inner limit based on index
    range0 = dates.index[0] + window
    range1 = dates.index[-1] - window
    # Outer limit based on valid points
    valid0 = selected.first_valid_index() + (window if inner_dates else null)
    valid1 = selected.last_valid_index() - (window if inner_dates else null)
    # Find inner limits of the two limits
    time0 = max(range0, valid0)
    time1 = min(range1, valid1)
    return time0, time1


def get_savgol_dates(dates, time0, time1, window):
    r"""Get start/end date ranges for Savitzky-Golay filter.

    Parameters
    ----------
    dates : Series
        Timestamps.
    time0, time1 : Timestamp
        Limit outside of which fittings will not be made.
    window : timedelta
        1/2 time window to perform the fitting.

    Returns
    -------
    date_s, date_e : Series
        Start time and end time of date ranges.
    """
    time0_series = pd.Series(time0, index=dates.index)
    time1_series = pd.Series(time1, index=dates.index)
    date_s = pd.Series(
        np.minimum(
            np.maximum(dates - window, time0_series),
            time1_series - 2*window
        ),
        index=dates.index
    )
    date_e = pd.Series(
        np.maximum(
            np.minimum(dates + window, time1_series),
            time0_series + 2*window
        ),
        index=dates.index
    )
    return date_s, date_e


def get_savgol_max_order(idata, col_in, window1, window2, inner_dates):
    r"""Get maximum order for Savitzky-Golay filter.

    Parameters
    ----------
    idata : DataFrame
        Input/Output data.
    col_in : str
        Input column.
    window1, window2 : timedelta
        1/2 time window to perform the fitting.
    inner_dates : boolean
        If true date limits are set to first/last valid index, else to extend
        by window size.

    Returns
    -------
    max_order : Series
        Maximum possible order for each data point.
    """
    selected = idata[col_in].dropna()
    dates = idata.index.to_series()
    window_total = window1 + window2
    time0, time1 = get_savgol_date_limits(
        idata=idata, col_in=col_in, window=window1, inner_dates=inner_dates
    )
    date_s, date_e = get_savgol_dates(
        dates=dates, time0=time0-window1, time1=time1+window1, window=window1
    )
    max_order = dates[time0 - window_total : time1 + window_total].apply(
        lambda x: np.array(selected.loc[date_s[x] : date_e[x]]).size - 1
    )
    return max_order


def savgol_coeffs(x, y, order):
    r"""Calculate fitted coefficents for Savitzky-Golay filter.
    Can work with non-continuous data.

    Parameters
    ----------
    x, y : array-like
        x and y values to perform the fit.
    order : int
        Order of the function.

    Raises
    ------
    Exception
        If x and y do not have the same shape.

    Returns
    -------
    value : ndarray (1d)
        Fitted coefficients.
    """
    if np.array(x).size == 0:
        return np.nan
    if np.array(x).shape != np.array(y).shape:
        raise Exception(
            f'Value Error: the shape of x {np.array(x).shape}'
            f'should be same as that of y {np.array(y).shape}'
        )
    # Calculate the design matrix with x
    A = np.array([x[t]**np.arange(order+1) for t in range(0, len(x), 1)])
    # Calculate inverse of (A.T x A)
    inv_tAA = np.linalg.pinv(np.matmul(A.T, A))
    # Calculate pseudoinverse of A
    coeffs = np.matmul(inv_tAA, A.T)
    # Calculate the fitted y at x=0
    value = np.array([(c * y).sum() for c in coeffs])
    return value


def savgol_value(x, y, order, x_i):
    r"""Calculate fitted coefficents for Savitzky-Golay filter.
    Can work with non-continuous data.

    Parameters
    ----------
    x, y : array-like
        x and y values to perform the fit.
    order : int
        Order of the function.

    Raises
    ------
    Exception
        If x and y do not have the same shape.

    Returns
    -------
    value : ndarray (1d)
        Fitted coefficients.

    """
    coeffs = savgol_coeffs(x, y, order)
    if x_i == 0:
        value = coeffs[0]
    else:
        value = sum([c * x_i**n for (n, c) in enumerate(coeffs)])
    return value